@@ -81,7 +81,7 @@ orc_test_gcc_compile (OrcProgram *p)
   fclose (file);
 
   file = fopen (dump_filename, "w");
-  ret = fwrite(p->code, p->code_size, 1, file);
+  ret = fwrite(p->orccode->code, p->orccode->code_size, 1, file);
   fclose (file);
 
 #if defined(HAVE_POWERPC)
@@ -190,7 +190,7 @@ orc_test_gcc_compile_neon (OrcProgram *p)
   fclose (file);
 
   file = fopen (dump_filename, "w");
-  ret = fwrite(p->code, p->code_size, 1, file);
+  ret = fwrite(p->orccode->code, p->orccode->code_size, 1, file);
   fclose (file);
 
   sprintf (cmd, PREFIX "gcc -march=armv6t2 -mcpu=cortex-a8 -mfpu=neon -Wall "
@@ -283,7 +283,7 @@ orc_test_gcc_compile_c64x (OrcProgram *p)
   fclose (file);
 
   file = fopen (dump_filename, "w");
-  ret = fwrite(p->code, p->code_size, 1, file);
+  ret = fwrite(p->orccode->code, p->orccode->code_size, 1, file);
   fclose (file);
 
   sprintf (cmd, C64X_PREFIX "cl6x -mv=6400+ "
@@ -359,7 +359,7 @@ orc_test_random_bits (void *data, int n_bytes)
 #endif
 }
 
-int
+static int
 print_array_val_signed (OrcArray *array, int i, int j)
 {
   void *ptr = ORC_PTR_OFFSET (array->data,
@@ -376,14 +376,16 @@ print_array_val_signed (OrcArray *array, int i, int j)
       printf(" %10d", *(orc_int32 *)ptr);
       return *(orc_int32 *)ptr;
     case 8:
-      printf(" %20lld", (long long)*(orc_int64 *)ptr);
+      printf(" 0x%08x%08x", (orc_uint32)((*(orc_uint64 *)ptr)>>32),
+          (orc_uint32)((*(orc_uint64 *)ptr)));
       return *(orc_int64 *)ptr;
     default:
       return -1;
   }
 }
 
-int
+#ifdef unused
+static int
 print_array_val_unsigned (OrcArray *array, int i, int j)
 {
   void *ptr = ORC_PTR_OFFSET (array->data,
@@ -400,14 +402,16 @@ print_array_val_unsigned (OrcArray *array, int i, int j)
       printf(" %10u", *(orc_uint32 *)ptr);
       return *(orc_int32 *)ptr;
     case 8:
-      printf(" %20llu", (long long)*(orc_uint64 *)ptr);
+      printf(" 0x%08x%08x", (orc_uint32)((*(orc_uint64 *)ptr)>>32),
+          (orc_uint32)((*(orc_uint64 *)ptr)));
       return *(orc_int64 *)ptr;
     default:
       return -1;
   }
 }
+#endif
 
-int
+static int
 print_array_val_hex (OrcArray *array, int i, int j)
 {
   void *ptr = ORC_PTR_OFFSET (array->data,
@@ -424,14 +428,15 @@ print_array_val_hex (OrcArray *array, int i, int j)
       printf(" %08x", *(orc_uint32 *)ptr);
       return *(orc_int32 *)ptr;
     case 8:
-      printf(" %016llx", (long long)*(orc_uint64 *)ptr);
+      printf(" 0x%08x%08x", (orc_uint32)((*(orc_uint64 *)ptr)>>32),
+          (orc_uint32)((*(orc_uint64 *)ptr)));
       return *(orc_int64 *)ptr;
     default:
       return -1;
   }
 }
 
-int
+static int
 print_array_val_float (OrcArray *array, int i, int j)
 {
   void *ptr = ORC_PTR_OFFSET (array->data,
@@ -521,10 +526,12 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
 
     result = orc_program_compile_full (program, target, flags);
     if (ORC_COMPILE_RESULT_IS_FATAL(result)) {
-      return ORC_TEST_FAILED;
+      ret = ORC_TEST_FAILED;
+      goto out;
     }
     if (!ORC_COMPILE_RESULT_IS_SUCCESSFUL(result)) {
-      return ORC_TEST_INDETERMINATE;
+      ret = ORC_TEST_INDETERMINATE;
+      goto out;
     }
   }
 
@@ -566,7 +573,20 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
       orc_array_set_pattern (dest_emul[i], ORC_OOB_VALUE);
       misalignment++;
     } else if (program->vars[i].vartype == ORC_VAR_TYPE_PARAM) {
-      orc_executor_set_param (ex, i, 2);
+      switch (program->vars[i].param_type) {
+        case ORC_PARAM_TYPE_INT:
+          orc_executor_set_param (ex, i, 2);
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          orc_executor_set_param_float (ex, i, 2.0);
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          orc_executor_set_param_int64 (ex, i, 2);
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          orc_executor_set_param_double (ex, i, 2.0);
+          break;
+      }
     }
   }
 
@@ -716,6 +736,9 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
 
   orc_executor_free (ex);
 
+out:
+  orc_program_reset (program);
+
   return ret;
 }
 
@@ -725,52 +748,9 @@ orc_test_get_program_for_opcode (OrcStaticOpcode *opcode)
   OrcProgram *p;
   char s[40];
   int flags = 0;
-
-  p = orc_program_new ();
-  if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    orc_program_add_accumulator (p, opcode->dest_size[0], "d1");
-  } else {
-    orc_program_add_destination (p, opcode->dest_size[0], "d1");
-  }
-  if (opcode->dest_size[1] != 0) {
-    orc_program_add_destination (p, opcode->dest_size[1], "d2");
-  }
-  orc_program_add_source (p, opcode->src_size[0], "s1");
-  if (opcode->src_size[1] != 0) {
-    orc_program_add_source (p, opcode->src_size[1], "s2");
-  }
-
-  if ((opcode->flags & ORC_STATIC_OPCODE_FLOAT_SRC) ||
-      (opcode->flags & ORC_STATIC_OPCODE_FLOAT_DEST)) {
-    flags = ORC_TEST_FLAGS_FLOAT;
-  }
-
-  sprintf(s, "test_%s", opcode->name);
-  orc_program_set_name (p, s);
-  orc_program_set_2d (p);
-
-  if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    orc_program_append_str (p, opcode->name, "a1", "s1", "s2");
-  } else {
-    orc_program_append_str (p, opcode->name, "d1", "s1", "s2");
-  }
-
-  return p;
-}
-
-OrcProgram *
-orc_test_get_program_for_opcode_const (OrcStaticOpcode *opcode)
-{
-  OrcProgram *p;
-  char s[40];
   int args[4] = { -1, -1, -1, -1 };
-  int flags;
   int n_args = 0;
 
-  if (opcode->src_size[1] == 0) {
-    return NULL;
-  }
-
   p = orc_program_new ();
   if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
     args[n_args++] =
@@ -783,13 +763,74 @@ orc_test_get_program_for_opcode_const (OrcStaticOpcode *opcode)
     args[n_args++] =
       orc_program_add_destination (p, opcode->dest_size[1], "d2");
   }
-  args[n_args++] =
-    orc_program_add_source (p, opcode->src_size[0], "s1");
-  args[n_args++] =
-    orc_program_add_constant (p, opcode->src_size[1], 1, "c1");
-  if (opcode->src_size[2]) {
+  if (opcode->flags & ORC_STATIC_OPCODE_SCALAR) {
+    if (opcode->src_size[1] == 0) {
+      args[n_args++] =
+        orc_program_add_constant (p, opcode->src_size[0], 1, "c1");
+    } else {
+      args[n_args++] =
+        orc_program_add_source (p, opcode->src_size[0], "s1");
+      args[n_args++] =
+        orc_program_add_constant (p, opcode->src_size[1], 1, "c1");
+      if (opcode->src_size[2] != 0) {
+        args[n_args++] =
+          orc_program_add_constant (p, opcode->src_size[1], 1, "c1");
+      }
+    }
+  } else {
     args[n_args++] =
-      orc_program_add_constant (p, opcode->src_size[2], 1, "c2");
+      orc_program_add_source (p, opcode->src_size[0], "s1");
+    args[n_args++] =
+      orc_program_add_source (p, opcode->src_size[1], "s2");
+  }
+
+  if ((opcode->flags & ORC_STATIC_OPCODE_FLOAT_SRC) ||
+      (opcode->flags & ORC_STATIC_OPCODE_FLOAT_DEST)) {
+    flags = ORC_TEST_FLAGS_FLOAT;
+  }
+
+  sprintf(s, "test_%s", opcode->name);
+  orc_program_set_name (p, s);
+
+  orc_program_append_2 (p, opcode->name, 0, args[0], args[1],
+      args[2], args[3]);
+
+  return p;
+}
+
+OrcProgram *
+orc_test_get_program_for_opcode_const (OrcStaticOpcode *opcode)
+{
+  OrcProgram *p;
+  char s[40];
+  int args[4] = { -1, -1, -1, -1 };
+  int flags;
+  int n_args = 0;
+
+  p = orc_program_new ();
+  if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
+    args[n_args++] =
+      orc_program_add_accumulator (p, opcode->dest_size[0], "d1");
+  } else {
+    args[n_args++] =
+      orc_program_add_destination (p, opcode->dest_size[0], "d1");
+  }
+  if (opcode->dest_size[1] != 0) {
+    args[n_args++] =
+      orc_program_add_destination (p, opcode->dest_size[1], "d2");
+  }
+  if (opcode->src_size[1] == 0) {
+    args[n_args++] =
+      orc_program_add_constant (p, opcode->src_size[0], 1, "c1");
+  } else {
+    args[n_args++] =
+      orc_program_add_source (p, opcode->src_size[0], "s1");
+    args[n_args++] =
+      orc_program_add_constant (p, opcode->src_size[1], 1, "c1");
+    if (opcode->src_size[2]) {
+      args[n_args++] =
+        orc_program_add_constant (p, opcode->src_size[2], 1, "c2");
+    }
   }
 
   if ((opcode->flags & ORC_STATIC_OPCODE_FLOAT_SRC) ||
@@ -889,6 +930,7 @@ orc_test_performance_full (OrcProgram *program, int flags,
     result = orc_program_compile_full (program, target, flags);
     if (!ORC_COMPILE_RESULT_IS_SUCCESSFUL(result)) {
       //printf("compile failed\n");
+      orc_program_reset (program);
       return 0;
     }
   }
@@ -951,14 +993,23 @@ orc_test_performance_full (OrcProgram *program, int flags,
         orc_executor_set_stride (ex, j, src[j-ORC_VAR_S1]->stride);
       }
     }
-    orc_profile_start (&prof);
-    orc_executor_run (ex);
-    orc_profile_stop (&prof);
+    if (flags & ORC_TEST_FLAGS_BACKUP) {
+      orc_profile_start (&prof);
+      orc_executor_run_backup (ex);
+      orc_profile_stop (&prof);
+    } else if (flags & ORC_TEST_FLAGS_EMULATE) {
+      orc_profile_start (&prof);
+      orc_executor_emulate (ex);
+      orc_profile_stop (&prof);
+    } else {
+      orc_profile_start (&prof);
+      orc_executor_run (ex);
+      orc_profile_stop (&prof);
+    }
   }
   ORC_DEBUG ("done running");
 
   orc_profile_get_ave_std (&prof, &ave, &std);
-  //printf("%g %g\n", ave/(n*m), std/(n*m));
 
   for(i=0;i<4;i++){
     if (dest_exec[i]) orc_array_free (dest_exec[i]);
@@ -969,6 +1020,7 @@ orc_test_performance_full (OrcProgram *program, int flags,
   }
 
   orc_executor_free (ex);
+  orc_program_reset (program);
 
   return ave/(n*m);
 }
