@@ -82,6 +82,8 @@ BINARY_SL(subssl, "ORC_CLAMP_SL((orc_int64)%s - (orc_int64)%s)")
 BINARY_UL(subusl, "ORC_CLAMP_UL((orc_int64)(orc_uint32)%s - (orc_int64)(orc_uint32)%s)")
 BINARY_SL(xorl, "%s ^ %s")
 
+UNARY_SQ(copyq, "%s")
+
 UNARY_BW(convsbw, "%s")
 UNARY_BW(convubw, "(orc_uint8)%s")
 UNARY_WB(convwb, "%s")
@@ -121,17 +123,22 @@ UNARY_LW(select0lw, "(orc_uint32)%s & 0xffff")
 UNARY_LW(select1lw, "((orc_uint32)%s >> 16)&0xffff")
 UNARY_UW(swapw, "ORC_SWAP_W(%s)")
 UNARY_UL(swapl, "ORC_SWAP_L(%s)")
+UNARY_UQ(swapq, "ORC_SWAP_Q(%s)")
 
-BINARY_F(addf, "ORC_DENORMAL(ORC_DENORMAL(%s) + ORC_DENORMAL(%s))")
-BINARY_F(subf, "ORC_DENORMAL(ORC_DENORMAL(%s) - ORC_DENORMAL(%s))")
-BINARY_F(mulf, "ORC_DENORMAL(ORC_DENORMAL(%s) * ORC_DENORMAL(%s))")
-BINARY_F(divf, "ORC_DENORMAL(ORC_DENORMAL(%s) / ORC_DENORMAL(%s))")
-UNARY_F(sqrtf, "sqrt(ORC_DENORMAL(%s))")
-BINARY_F(maxf, "ORC_MAXF(ORC_DENORMAL(%s),ORC_DENORMAL(%s))")
-BINARY_F(minf, "ORC_MINF(ORC_DENORMAL(%s),ORC_DENORMAL(%s))")
-BINARY_FL(cmpeqf, "(ORC_DENORMAL(%s) == ORC_DENORMAL(%s)) ? (~0) : 0")
-BINARY_FL(cmpltf, "(ORC_DENORMAL(%s) < ORC_DENORMAL(%s)) ? (~0) : 0")
-BINARY_FL(cmplef, "(ORC_DENORMAL(%s) <= ORC_DENORMAL(%s)) ? (~0) : 0")
-UNARY_FL(convfl, "(int)%s")
-UNARY_LF(convlf, "%s")
+BINARY_F(addf, "%s + %s")
+BINARY_F(subf, "%s - %s")
+BINARY_F(mulf, "%s * %s")
+BINARY_F(divf, "%s / %s")
+UNARY_F(sqrtf, "sqrt(%s)")
+BINARY_FL(cmpeqf, "(%s == %s) ? (~0) : 0")
+BINARY_FL(cmpltf, "(%s < %s) ? (~0) : 0")
+BINARY_FL(cmplef, "(%s <= %s) ? (~0) : 0")
 
+BINARY_D(addd, "%s + %s")
+BINARY_D(subd, "%s - %s")
+BINARY_D(muld, "%s * %s")
+BINARY_D(divd, "%s / %s")
+UNARY_D(sqrtd, "sqrt(%s)")
+BINARY_DQ(cmpeqd, "(%s == %s) ? (~0ULL) : 0")
+BINARY_DQ(cmpltd, "(%s < %s) ? (~0ULL) : 0")
+BINARY_DQ(cmpled, "(%s <= %s) ? (~0ULL) : 0")
