@@ -278,6 +278,7 @@ void
 emulate_null (OrcOpcodeExecutor *ex, int offset, int n)
 {
   /* This is a placeholder for adding new opcodes */
+  ORC_ERROR("emulate_null() called.  This is a bug.");
 }
 
 #include "orc/orcemulateopcodes.h"
@@ -395,6 +396,7 @@ static OrcStaticOpcode opcodes[] = {
   { "loadpq", ORC_STATIC_OPCODE_LOAD|ORC_STATIC_OPCODE_SCALAR|ORC_STATIC_OPCODE_INVARIANT, { 8 }, { 8 }, emulate_loadpq },
   { "storeq", ORC_STATIC_OPCODE_STORE, { 8 }, { 8 }, emulate_storeq },
   { "splatw3q", 0, { 8 }, { 8 }, emulate_splatw3q },
+  { "copyq", 0, { 8 }, { 8 }, emulate_copyq },
 
   { "convsbw", 0, { 2 }, { 1 }, emulate_convsbw },
   { "convubw", 0, { 2 }, { 1 }, emulate_convubw },
@@ -442,12 +444,14 @@ static OrcStaticOpcode opcodes[] = {
 
   { "swapw", 0, { 2 }, { 2 }, emulate_swapw },
   { "swapl", 0, { 4 }, { 4 }, emulate_swapl },
+  { "swapq", 0, { 8 }, { 8 }, emulate_swapq },
   { "select0wb", 0, { 1 }, { 2 }, emulate_select0wb },
   { "select1wb", 0, { 1 }, { 2 }, emulate_select1wb },
   { "select0lw", 0, { 2 }, { 4 }, emulate_select0lw },
   { "select1lw", 0, { 2 }, { 4 }, emulate_select1lw },
   { "mergewl", 0, { 4 }, { 2, 2 }, emulate_mergewl },
   { "mergebw", 0, { 2 }, { 1, 1 }, emulate_mergebw },
+  { "splitql", 0, { 4, 4 }, { 8 }, emulate_splitql },
   { "splitlw", 0, { 2, 2 }, { 4 }, emulate_splitlw },
   { "splitwb", 0, { 1, 1 }, { 2 }, emulate_splitwb },
 
@@ -465,6 +469,22 @@ static OrcStaticOpcode opcodes[] = {
   { "convfl", ORC_STATIC_OPCODE_FLOAT_SRC, { 4 }, { 4 }, emulate_convfl },
   { "convlf", ORC_STATIC_OPCODE_FLOAT_DEST, { 4 }, { 4 }, emulate_convlf },
 
+  /* double ops */
+  { "addd", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8, 8 }, emulate_addd },
+  { "subd", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8, 8 }, emulate_subd },
+  { "muld", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8, 8 }, emulate_muld },
+  { "divd", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8, 8 }, emulate_divd },
+  { "sqrtd", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8 }, emulate_sqrtd },
+  { "maxd", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8, 8 }, emulate_maxd },
+  { "mind", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 8, 8 }, emulate_mind },
+  { "cmpeqd", ORC_STATIC_OPCODE_FLOAT_SRC, { 8 }, { 8, 8 }, emulate_cmpeqd },
+  { "cmpltd", ORC_STATIC_OPCODE_FLOAT_SRC, { 8 }, { 8, 8 }, emulate_cmpltd },
+  { "cmpled", ORC_STATIC_OPCODE_FLOAT_SRC, { 8 }, { 8, 8 }, emulate_cmpled },
+  { "convdl", ORC_STATIC_OPCODE_FLOAT_SRC, { 4 }, { 8 }, emulate_convdl },
+  { "convld", ORC_STATIC_OPCODE_FLOAT_DEST, { 8 }, { 4 }, emulate_convld },
+  { "convfd", ORC_STATIC_OPCODE_FLOAT, { 8 }, { 4 }, emulate_convfd },
+  { "convdf", ORC_STATIC_OPCODE_FLOAT, { 4 }, { 8 }, emulate_convdf },
+
   { "" }
 };
 
