@@ -186,8 +186,8 @@ orc_parse_full (const char *code, OrcProgram ***programs, char **log)
         orc_program_add_parameter (parser->program, size, token[2]);
       } else if (strcmp (token[0], ".const") == 0) {
         int size = strtol (token[1], NULL, 0);
-        int value = strtoul (token[3], NULL, 0);
-        orc_program_add_constant (parser->program, size, value, token[2]);
+
+        orc_program_add_constant_str (parser->program, size, token[3], token[2]);
       } else if (strcmp (token[0], ".floatparam") == 0) {
         int size = strtol (token[1], NULL, 0);
         orc_program_add_parameter_float (parser->program, size, token[2]);
@@ -212,6 +212,7 @@ orc_parse_full (const char *code, OrcProgram ***programs, char **log)
 
       if (o) {
         int n_args = opcode_n_args (o);
+        int i;
 
         if (n_tokens != 1 + offset + n_args) {
           orc_parse_log (parser, "error: line %d: too %s arguments for %s (expected %d)\n",
@@ -219,23 +220,22 @@ orc_parse_full (const char *code, OrcProgram ***programs, char **log)
               token[offset], n_args);
         }
 
+        for(i=offset+1;i<n_tokens;i++){
+          char *end;
+          double d;
+          d = strtod (token[i], &end);
+          if (end != token[i]) {
+            orc_program_add_constant_str (parser->program, 0, token[i],
+                token[i]);
+          }
+        }
+
         if (n_tokens - offset == 5) {
           orc_program_append_str_2 (parser->program, token[offset], flags,
               token[offset+1], token[offset+2], token[offset+3], token[offset+4]);
         } else if (n_tokens - offset == 4) {
-          char *end;
-          int imm = strtol (token[offset + 3], &end, 0);
-          if (end != token[offset + 3]) {
-            char creg[10];
-            sprintf(creg, "c%d", parser->creg_index);
-            parser->creg_index++;
-            orc_program_add_constant (parser->program, 2, imm, creg);
-            orc_program_append_str_2 (parser->program, token[offset], flags,
-                token[offset+1], token[offset+2], creg, NULL);
-          } else {
-            orc_program_append_str_2 (parser->program, token[offset], flags,
-                token[offset+1], token[offset+2], token[offset+3], NULL);
-          }
+          orc_program_append_str_2 (parser->program, token[offset], flags,
+              token[offset+1], token[offset+2], token[offset+3], NULL);
         } else {
           orc_program_append_str_2 (parser->program, token[offset], flags,
               token[offset+1], token[offset+2], NULL, NULL);
@@ -259,7 +259,9 @@ orc_parse_full (const char *code, OrcProgram ***programs, char **log)
   } else {
     free (parser->log);
   }
-  parser->programs[0]->init_function = init_function;
+  if (parser->programs[0]) {
+    parser->programs[0]->init_function = init_function;
+  }
   *programs = parser->programs;
   return parser->n_programs;
 }
@@ -302,7 +304,7 @@ orc_parse_log_valist (OrcParser *parser, const char *format, va_list args)
     sprintf(s, "In function %s:\n", parser->program->name);
     len = strlen(s);
 
-    if (parser->log_size + len > parser->log_alloc) {
+    if (parser->log_size + len + 1 >= parser->log_alloc) {
       parser->log_alloc += 100;
       parser->log = realloc (parser->log, parser->log_alloc);
     }
@@ -315,7 +317,7 @@ orc_parse_log_valist (OrcParser *parser, const char *format, va_list args)
   vsprintf(s, format, args);
   len = strlen(s);
 
-  if (parser->log_size + len > parser->log_alloc) {
+  if (parser->log_size + len + 1 >= parser->log_alloc) {
     parser->log_alloc += 100;
     parser->log = realloc (parser->log, parser->log_alloc);
   }
