@@ -19,7 +19,7 @@ orc_compiler_c_init (OrcCompiler *compiler)
 {
   int i;
 
-  for(i=ORC_GP_REG_BASE;i<ORC_GP_REG_BASE+16;i++){
+  for(i=ORC_GP_REG_BASE;i<ORC_GP_REG_BASE+32;i++){
     compiler->valid_regs[i] = 1;
   }
   compiler->loop_shift = 0;
@@ -101,11 +101,19 @@ orc_target_c_get_asm_preamble (void)
     "#define ORC_CLAMP_UL(x) ORC_CLAMP(x,ORC_UL_MIN,ORC_UL_MAX)\n"
     "#define ORC_SWAP_W(x) ((((x)&0xff)<<8) | (((x)&0xff00)>>8))\n"
     "#define ORC_SWAP_L(x) ((((x)&0xff)<<24) | (((x)&0xff00)<<8) | (((x)&0xff0000)>>8) | (((x)&0xff000000)>>24))\n"
+    "#define ORC_SWAP_Q(x) ((((x)&0xffULL)<<56) | (((x)&0xff00ULL)<<40) | (((x)&0xff0000ULL)<<24) | (((x)&0xff000000ULL)<<8) | (((x)&0xff00000000ULL)>>8) | (((x)&0xff0000000000ULL)>>24) | (((x)&0xff000000000000ULL)>>40) | (((x)&0xff00000000000000ULL)>>56))\n"
     "#define ORC_PTR_OFFSET(ptr,offset) ((void *)(((unsigned char *)(ptr)) + (offset)))\n"
-    "#define ORC_MIN_NORMAL (1.1754944909521339405e-38)\n"
-    "#define ORC_DENORMAL(x) (((x) > -ORC_MIN_NORMAL && (x) < ORC_MIN_NORMAL) ? ((x)<0 ? (-0.0f) : (0.0f)) : (x))\n"
-    "#define ORC_MINF(a,b) (isnan(a) ? a : isnan(b) ? b : ((a)<(b)) ? (a) : (b))\n"
-    "#define ORC_MAXF(a,b) (isnan(a) ? a : isnan(b) ? b : ((a)>(b)) ? (a) : (b))\n"
+    "#define ORC_DENORMAL(x) ((x) & ((((x)&0x7f800000) == 0) ? 0xff800000 : 0xffffffff))\n"
+    "#define ORC_ISNAN(x) ((((x)&0x7f800000) == 0x7f800000) && (((x)&0x007fffff) != 0))\n"
+    "#define ORC_DENORMAL_DOUBLE(x) ((x) & ((((x)&0x7ff0000000000000ULL) == 0) ? 0xfff0000000000000ULL : 0xffffffffffffffffULL))\n"
+    "#define ORC_ISNAN_DOUBLE(x) ((((x)&0x7ff0000000000000ULL) == 0x7ff0000000000000ULL) && (((x)&0x000fffffffffffffULL) != 0))\n"
+    "#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L\n"
+    "#define ORC_RESTRICT restrict\n"
+    "#elif defined(__GNUC__) && __GNUC__ >= 4\n"
+    "#define ORC_RESTRICT __restrict__\n"
+    "#else\n"
+    "#define ORC_RESTRICT\n"
+    "#endif\n"
     "/* end Orc C target preamble */\n\n";
 }
 
@@ -210,15 +218,13 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
         }
         break;
       case ORC_VAR_TYPE_SRC:
-        ORC_ASM_CODE(compiler,"  const %s *%s ptr%d;\n",
+        ORC_ASM_CODE(compiler,"  const %s * ORC_RESTRICT ptr%d;\n",
             c_get_type_name (var->size),
-            (compiler->target_flags & ORC_TARGET_C_C99) ? "restrict " : "",
             i);
         break;
       case ORC_VAR_TYPE_DEST:
-        ORC_ASM_CODE(compiler,"  %s *%s ptr%d;\n",
+        ORC_ASM_CODE(compiler,"  %s * ORC_RESTRICT ptr%d;\n",
             c_get_type_name (var->size),
-            (compiler->target_flags & ORC_TARGET_C_C99) ? "restrict " : "",
             i);
         break;
       case ORC_VAR_TYPE_ACCUMULATOR:
@@ -294,6 +300,41 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
     }
   }
 
+  ORC_ASM_CODE(compiler,"\n");
+  for(j=0;j<compiler->n_insns;j++){
+    insn = compiler->insns + j;
+    opcode = insn->opcode;
+
+    if (!(insn->flags & ORC_INSN_FLAG_INVARIANT)) continue;
+
+    ORC_ASM_CODE(compiler,"%*s    /* %d: %s */\n", prefix, "",
+        j, insn->opcode->name);
+
+    rule = insn->rule;
+    if (!rule) {
+      ORC_COMPILER_ERROR(compiler, "No rule for: %s on target %s", opcode->name,
+          compiler->target->name);
+      compiler->error = TRUE;
+    }
+    ORC_ASM_CODE(compiler,"%*s", prefix, "");
+    if (insn->flags & (ORC_INSTRUCTION_FLAG_X2|ORC_INSTRUCTION_FLAG_X4)) {
+      int n;
+      if (insn->flags & ORC_INSTRUCTION_FLAG_X2) {
+        n = 2;
+      } else {
+        n = 4;
+      }
+      for(i=0;i<n;i++){
+        compiler->unroll_index = i;
+        ORC_ASM_CODE(compiler,"%*s", prefix, "");
+        rule->emit (compiler, rule->emit_user, insn);
+      }
+    } else {
+      ORC_ASM_CODE(compiler,"%*s", prefix, "");
+      rule->emit (compiler, rule->emit_user, insn);
+    }
+  }
+
   ORC_ASM_CODE(compiler,"\n");
   ORC_ASM_CODE(compiler,"%*s  for (i = 0; i < n; i++) {\n", prefix, "");
 
@@ -302,6 +343,8 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
     insn = compiler->insns + j;
     opcode = insn->opcode;
 
+    if (insn->flags & ORC_INSN_FLAG_INVARIANT) continue;
+
     ORC_ASM_CODE(compiler,"%*s    /* %d: %s */\n", prefix, "",
         j, insn->opcode->name);
 
@@ -338,9 +381,9 @@ orc_compiler_c_assemble (OrcCompiler *compiler)
     char varname[40];
     OrcVariable *var = compiler->vars + i;
     if (var->name == NULL) continue;
-    c_get_name_int (varname, compiler, NULL, i);
     switch (var->vartype) {
       case ORC_VAR_TYPE_ACCUMULATOR:
+        c_get_name_int (varname, compiler, NULL, i);
         if (var->size == 2) {
           if (compiler->target_flags & ORC_TARGET_C_NOEXEC) {
             ORC_ASM_CODE(compiler,"  *%s = (%s & 0xffff);\n",
@@ -387,25 +430,55 @@ c_get_name_int (char *name, OrcCompiler *p, OrcInstruction *insn, int var)
     if (p->target_flags & ORC_TARGET_C_NOEXEC) {
       sprintf(name,"%s", varnames[var]);
     } else if (p->target_flags & ORC_TARGET_C_OPCODE) {
-      if (p->vars[var].is_float_param) {
-        sprintf(name,"((orc_union32 *)(ex->src_ptrs[%d]))->f",
-            var - ORC_VAR_P1 + p->program->n_src_vars);
-      } else {
-        sprintf(name,"((orc_union32 *)(ex->src_ptrs[%d]))->i",
-            var - ORC_VAR_P1 + p->program->n_src_vars);
+      switch (p->vars[var].param_type) {
+        case ORC_PARAM_TYPE_INT:
+          sprintf(name,"((orc_union32 *)(ex->src_ptrs[%d]))->i",
+              var - ORC_VAR_P1 + p->program->n_src_vars);
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          sprintf(name,"((orc_union32 *)(ex->src_ptrs[%d]))->i",
+              var - ORC_VAR_P1 + p->program->n_src_vars);
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          sprintf(name,"((orc_union64 *)(ex->src_ptrs[%d]))->i",
+              var - ORC_VAR_P1 + p->program->n_src_vars);
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          sprintf(name,"((orc_union64 *)(ex->src_ptrs[%d]))->i",
+              var - ORC_VAR_P1 + p->program->n_src_vars);
+          break;
+        default:
+          ORC_ASSERT(0);
       }
     } else {
-      if (p->vars[var].is_float_param) {
-        sprintf(name,"((orc_union32 *)(ex->params+%d))->f", var);
-      } else {
-        sprintf(name,"ex->params[%d]", var);
+      switch (p->vars[var].param_type) {
+        case ORC_PARAM_TYPE_INT:
+          sprintf(name,"ex->params[%d]", var);
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          sprintf(name,"((orc_union32 *)(ex->params+%d))->i", var);
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          /* FIXME */
+          sprintf(name,"((orc_union32 *)(ex->params+%d))->i", var);
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          /* FIXME */
+          sprintf(name,"((orc_union32 *)(ex->params+%d))->i", var);
+          break;
+        default:
+          ORC_ASSERT(0);
       }
     }
   } else if (p->vars[var].vartype == ORC_VAR_TYPE_CONST) {
-    if (p->vars[var].value == 0x80000000) {
+    if (p->vars[var].value.i == 0x80000000) {
       sprintf(name,"0x80000000");
     } else {
-      sprintf(name, "%d", p->vars[var].value);
+      if (p->vars[var].value.i == (int)p->vars[var].value.i) {
+        sprintf(name, "%d", (int)p->vars[var].value.i);
+      } else {
+        ORC_ASSERT(0);
+      }
     }
   } else {
     if (insn && (insn->flags & ORC_INSTRUCTION_FLAG_X2)) {
@@ -498,10 +571,16 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 { \
   char dest[40], src1[40]; \
 \
-  c_get_name_float (dest, p, insn, insn->dest_args[0]); \
-  c_get_name_float (src1, p, insn, insn->src_args[0]); \
+  c_get_name_int (dest, p, insn, insn->dest_args[0]); \
+  c_get_name_int (src1, p, insn, insn->src_args[0]); \
  \
-  ORC_ASM_CODE(p,"    %s = " op ";\n", dest, src1); \
+  ORC_ASM_CODE(p, "    {\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _src1;\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _dest1;\n"); \
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL(%s);\n", src1); \
+  ORC_ASM_CODE(p,"       _dest1.f = " op ";\n", "_src1.f"); \
+  ORC_ASM_CODE(p,"       %s = ORC_DENORMAL(_dest1.i);\n", dest); \
+  ORC_ASM_CODE(p, "    }\n"); \
 }
 
 #define BINARYF(name,op) \
@@ -510,11 +589,19 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 { \
   char dest[40], src1[40], src2[40]; \
 \
-  c_get_name_float (dest, p, insn, insn->dest_args[0]); \
-  c_get_name_float (src1, p, insn, insn->src_args[0]); \
-  c_get_name_float (src2, p, insn, insn->src_args[1]); \
+  c_get_name_int (dest, p, insn, insn->dest_args[0]); \
+  c_get_name_int (src1, p, insn, insn->src_args[0]); \
+  c_get_name_int (src2, p, insn, insn->src_args[1]); \
  \
-  ORC_ASM_CODE(p,"    %s = " op ";\n", dest, src1, src2); \
+  ORC_ASM_CODE(p, "    {\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _src1;\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _src2;\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _dest1;\n"); \
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL(%s);\n", src1); \
+  ORC_ASM_CODE(p,"       _src2.i = ORC_DENORMAL(%s);\n", src2); \
+  ORC_ASM_CODE(p,"       _dest1.f = " op ";\n", "_src1.f", "_src2.f"); \
+  ORC_ASM_CODE(p,"       %s = ORC_DENORMAL(_dest1.i);\n", dest); \
+  ORC_ASM_CODE(p, "    }\n"); \
 }
 
 #define BINARYFL(name,op) \
@@ -524,34 +611,74 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
   char dest[40], src1[40], src2[40]; \
 \
   c_get_name_int (dest, p, insn, insn->dest_args[0]); \
-  c_get_name_float (src1, p, insn, insn->src_args[0]); \
-  c_get_name_float (src2, p, insn, insn->src_args[1]); \
+  c_get_name_int (src1, p, insn, insn->src_args[0]); \
+  c_get_name_int (src2, p, insn, insn->src_args[1]); \
  \
-  ORC_ASM_CODE(p,"    %s = " op ";\n", dest, src1, src2); \
+  ORC_ASM_CODE(p, "    {\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _src1;\n"); \
+  ORC_ASM_CODE(p,"       orc_union32 _src2;\n"); \
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL(%s);\n", src1); \
+  ORC_ASM_CODE(p,"       _src2.i = ORC_DENORMAL(%s);\n", src2); \
+  ORC_ASM_CODE(p,"       %s = " op ";\n", dest, "_src1.f", "_src2.f"); \
+  ORC_ASM_CODE(p, "    }\n"); \
 }
 
-#define UNARYFL(name,op) \
+#define UNARYD(name,op) \
 static void \
 c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 { \
   char dest[40], src1[40]; \
 \
   c_get_name_int (dest, p, insn, insn->dest_args[0]); \
-  c_get_name_float (src1, p, insn, insn->src_args[0]); \
+  c_get_name_int (src1, p, insn, insn->src_args[0]); \
  \
-  ORC_ASM_CODE(p,"    %s = " op ";\n", dest, src1); \
+  ORC_ASM_CODE(p, "    {\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _src1;\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _dest1;\n"); \
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL_DOUBLE(%s);\n", src1); \
+  ORC_ASM_CODE(p,"       _dest1.f = " op ";\n", "_src1.f"); \
+  ORC_ASM_CODE(p,"       %s = ORC_DENORMAL_DOUBLE(_dest1.i);\n", dest); \
+  ORC_ASM_CODE(p, "    }\n"); \
 }
 
-#define UNARYLF(name,op) \
+#define BINARYD(name,op) \
 static void \
 c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 { \
-  char dest[40], src1[40]; \
+  char dest[40], src1[40], src2[40]; \
 \
-  c_get_name_float (dest, p, insn, insn->dest_args[0]); \
+  c_get_name_int (dest, p, insn, insn->dest_args[0]); \
   c_get_name_int (src1, p, insn, insn->src_args[0]); \
+  c_get_name_int (src2, p, insn, insn->src_args[1]); \
  \
-  ORC_ASM_CODE(p,"    %s = " op ";\n", dest, src1); \
+  ORC_ASM_CODE(p, "    {\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _src1;\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _src2;\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _dest1;\n"); \
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL_DOUBLE(%s);\n", src1); \
+  ORC_ASM_CODE(p,"       _src2.i = ORC_DENORMAL_DOUBLE(%s);\n", src2); \
+  ORC_ASM_CODE(p,"       _dest1.f = " op ";\n", "_src1.f", "_src2.f"); \
+  ORC_ASM_CODE(p,"       %s = ORC_DENORMAL_DOUBLE(_dest1.i);\n", dest); \
+  ORC_ASM_CODE(p, "    }\n"); \
+}
+
+#define BINARYDQ(name,op) \
+static void \
+c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
+{ \
+  char dest[40], src1[40], src2[40]; \
+\
+  c_get_name_int (dest, p, insn, insn->dest_args[0]); \
+  c_get_name_int (src1, p, insn, insn->src_args[0]); \
+  c_get_name_int (src2, p, insn, insn->src_args[1]); \
+ \
+  ORC_ASM_CODE(p, "    {\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _src1;\n"); \
+  ORC_ASM_CODE(p,"       orc_union64 _src2;\n"); \
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL_DOUBLE(%s);\n", src1); \
+  ORC_ASM_CODE(p,"       _src2.i = ORC_DENORMAL_DOUBLE(%s);\n", src2); \
+  ORC_ASM_CODE(p,"       %s = " op ";\n", dest, "_src1.f", "_src2.f"); \
+  ORC_ASM_CODE(p, "    }\n"); \
 }
 
 #define BINARY_SB(a,b) BINARY(a,b)
@@ -587,6 +714,14 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 #define UNARY_FL(a,b) UNARYFL(a,b)
 #define UNARY_LF(a,b) UNARYLF(a,b)
 
+#define BINARY_D(a,b) BINARYD(a,b)
+#define BINARY_DQ(a,b) BINARYDQ(a,b)
+#define UNARY_D(a,b) UNARYD(a,b)
+#define UNARY_DL(a,b) UNARYFL(a,b)
+#define UNARY_LD(a,b) UNARYLF(a,b)
+#define UNARY_DF(a,b) UNARYF(a,b)
+#define UNARY_FD(a,b) UNARYF(a,b)
+
 #include "opcodes.h"
 
 #undef BINARY_SB
@@ -598,6 +733,7 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 #undef BINARY_SQ
 #undef BINARY_UQ
 #undef BINARY_F
+#undef BINARY_D
 #undef UNARY_SB
 #undef UNARY_UB
 #undef UNARY_SW
@@ -607,6 +743,7 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 #undef UNARY_SQ
 #undef UNARY_UQ
 #undef UNARY_F
+#undef UNARY_D
 #undef BINARY_BW
 #undef BINARY_WL
 #undef BINARY_LW
@@ -618,9 +755,61 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 #undef UNARY_LW
 #undef UNARY_WB
 #undef UNARY_FL
+#undef UNARY_DL
 #undef UNARY_LF
+#undef UNARY_LD
 #undef BINARY_FL
+#undef BINARY_DQ
+#undef UNARY_FD
+#undef UNARY_DF
 
+
+static void
+c_rule_loadpX (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40];
+  int size = ORC_PTR_TO_INT(user);
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+
+  if (p->vars[insn->src_args[0]].vartype == ORC_VAR_TYPE_PARAM) {
+    if (p->target_flags & ORC_TARGET_C_NOEXEC) {
+      ORC_ASM_CODE(p,"    %s = %s;\n", dest, varnames[insn->src_args[0]]);
+    } else if (p->target_flags & ORC_TARGET_C_OPCODE) {
+      if (size == 8) {
+        ORC_ASM_CODE(p,"    %s = ((orc_union64 *)(ex->src_ptrs[%d]))->i;\n",
+            dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
+      } else {
+        ORC_ASM_CODE(p,"    %s = ((orc_union32 *)(ex->src_ptrs[%d]))->i;\n",
+            dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
+      }
+    } else {
+      if (size == 8) {
+        ORC_ASM_CODE(p,"    %s = (ex->params[%d] & 0xffffffff) | (((orc_uint64)(ex->params[%d + (ORC_VAR_T1 - ORC_VAR_P1)]) << 32);\n",
+            dest, insn->src_args[0], insn->src_args[0]);
+      } else {
+        ORC_ASM_CODE(p,"    %s = ex->params[%d];\n", dest,
+            insn->src_args[0]);
+      }
+    }
+  } else if (p->vars[insn->src_args[0]].vartype == ORC_VAR_TYPE_CONST) {
+    if (p->vars[insn->src_args[0]].size <= 4) {
+      ORC_ASM_CODE(p,"    %s = 0x%08x; /* %d or %gf */\n", dest,
+          (unsigned int)p->vars[insn->src_args[0]].value.i,
+          (int)p->vars[insn->src_args[0]].value.i,
+          p->vars[insn->src_args[0]].value.f);
+    } else {
+      ORC_ASM_CODE(p,"    %s = 0x%08x%08x; /* %gf */\n", dest,
+          (orc_uint32)(((orc_uint64)p->vars[insn->src_args[0]].value.i)>>32),
+          ((orc_uint32)p->vars[insn->src_args[0]].value.i),
+          p->vars[insn->src_args[0]].value.f);
+    }
+  } else {
+    ORC_COMPILER_ERROR(p, "expected param or constant");
+  }
+}
+
+#if 0
 static void
 c_rule_loadpX (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
@@ -634,37 +823,8 @@ c_rule_loadpX (OrcCompiler *p, void *user, OrcInstruction *insn)
   var = &p->vars[insn->src_args[0]];
 
   ORC_ASM_CODE(p,"    %s = %s;\n", dest, src);
-#if 0
-  if (var->vartype == ORC_VAR_TYPE_CONST) {
-    if (var->value == 0x80000000) {
-      ORC_ASM_CODE(p,"    %s = 0x80000000;\n", dest);
-    } else {
-      ORC_ASM_CODE(p,"    %s = %d;\n", dest, var->value);
-    }
-  } else {
-    if (p->target_flags & ORC_TARGET_C_NOEXEC) {
-      ORC_ASM_CODE(p,"  %s = %s;\n", dest, varnames[insn->src_args[0]]);
-    } else if (p->target_flags & ORC_TARGET_C_OPCODE) {
-      if (var->is_float_param) {
-        c_get_name_float (dest, p, insn, insn->dest_args[0]);
-        ORC_ASM_CODE(p,"  %s = ((orc_union32 *)(ex->src_ptrs[%d]))->f;\n",
-            dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
-      } else {
-        ORC_ASM_CODE(p,"  %s = ((orc_union32 *)(ex->src_ptrs[%d]))->i;\n",
-            dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
-      }
-    } else {
-      if (var->is_float_param) {
-        ORC_ASM_CODE(p,"  %s = ((orc_union32 *)(ex->params+%d))->f;\n",
-            dest, insn->src_args[0]);
-      } else {
-        ORC_ASM_CODE(p,"  %s = ex->params[%d];\n",
-            dest, insn->src_args[0]);
-      }
-    }
-  }
-#endif
 }
+#endif
 
 static void
 c_rule_loadX (OrcCompiler *p, void *user, OrcInstruction *insn)
@@ -785,7 +945,7 @@ c_rule_ldreslinl (OrcCompiler *p, void *user, OrcInstruction *insn)
   ORC_ASM_CODE(p,"    orc_union32 a = ptr%d[tmp>>16];\n", insn->src_args[0]);
   ORC_ASM_CODE(p,"    orc_union32 b = ptr%d[(tmp>>16)+1];\n", insn->src_args[0]);
   for (i=0;i<4;i++){
-    ORC_ASM_CODE(p,"    var%d.x4[%d] = a.x4[%d] * (256-((tmp>>8)&0xff)) + b.x4[%d] * ((tmp>>8)&0xff);\n",
+    ORC_ASM_CODE(p,"    var%d.x4[%d] = ((orc_uint8)a.x4[%d] * (256-((tmp>>8)&0xff)) + (orc_uint8)b.x4[%d] * ((tmp>>8)&0xff))>>8;\n",
         insn->dest_args[0], i, i, i);
   }
   ORC_ASM_CODE(p,"    }\n");
@@ -840,6 +1000,19 @@ c_rule_accsadubl (OrcCompiler *p, void *user, OrcInstruction *insn)
       dest, dest, src1, src2);
 }
 
+static void
+c_rule_splitql (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest1[40], dest2[40], src[40];
+
+  c_get_name_int (dest1, p, insn, insn->dest_args[0]);
+  c_get_name_int (dest2, p, insn, insn->dest_args[1]);
+  c_get_name_int (src, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p,"    %s = (%s >> 32) & 0xffffffff;\n", dest1, src);
+  ORC_ASM_CODE(p,"    %s = %s & 0xffffffff;\n", dest2, src);
+}
+
 static void
 c_rule_splitlw (OrcCompiler *p, void *user, OrcInstruction *insn)
 {
@@ -933,6 +1106,174 @@ c_rule_divluw (OrcCompiler *p, void *user, OrcInstruction *insn)
       dest, src2, src1, src2);
 }
 
+static void
+c_rule_convlf (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40];
+
+  c_get_name_float (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p,"     %s = %s;\n", dest, src1);
+}
+
+static void
+c_rule_convld (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40];
+
+  c_get_name_float (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p,"     %s = %s;\n", dest, src1);
+}
+
+static void
+c_rule_convfd (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40];
+
+  c_get_name_float (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p, "    {\n");
+  ORC_ASM_CODE(p,"       orc_union32 _src1;\n");
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL(%s);\n", src1);
+  ORC_ASM_CODE(p,"       %s = _src1.f;\n", dest);
+  ORC_ASM_CODE(p, "    }\n");
+}
+
+static void
+c_rule_convdf (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p, "    {\n");
+  ORC_ASM_CODE(p,"       orc_union64 _src1;\n");
+  ORC_ASM_CODE(p,"       orc_union32 _dest;\n");
+  ORC_ASM_CODE(p,"       _src1.i = ORC_DENORMAL_DOUBLE(%s);\n", src1);
+  ORC_ASM_CODE(p,"       _dest.f = _src1.f;\n");
+  ORC_ASM_CODE(p,"       %s = ORC_DENORMAL(_dest.i);\n", dest);
+  ORC_ASM_CODE(p, "    }\n");
+}
+
+static void
+c_rule_convfl (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src[40], src_i[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_float (src, p, insn, insn->src_args[0]);
+  c_get_name_int (src_i, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p, "    {\n");
+  ORC_ASM_CODE(p,"       int tmp;\n");
+  ORC_ASM_CODE(p,"       tmp = (int)%s;\n", src);
+  ORC_ASM_CODE(p,"       if (tmp == 0x80000000 && !(%s&0x80000000)) tmp = 0x7fffffff;\n", src_i);
+  ORC_ASM_CODE(p,"       %s = tmp;\n", dest);
+  ORC_ASM_CODE(p, "    }\n");
+}
+
+static void
+c_rule_convdl (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src[40], src_i[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_float (src, p, insn, insn->src_args[0]);
+  c_get_name_int (src_i, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p, "    {\n");
+  ORC_ASM_CODE(p,"       int tmp;\n");
+  ORC_ASM_CODE(p,"       tmp = %s;\n", src);
+  ORC_ASM_CODE(p,"       if (tmp == 0x80000000 && !(%s&0x8000000000000000ULL)) tmp = 0x7fffffff;\n", src_i);
+  ORC_ASM_CODE(p,"       %s = tmp;\n", dest);
+  ORC_ASM_CODE(p, "    }\n");
+}
+
+static void
+c_rule_minf (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40], src2[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+  c_get_name_int (src2, p, insn, insn->src_args[1]);
+
+  ORC_ASM_CODE(p,"    {\n");
+  ORC_ASM_CODE(p,"      orc_union32 _src1;\n");
+  ORC_ASM_CODE(p,"      orc_union32 _src2;\n");
+  ORC_ASM_CODE(p,"      _src1.i = ORC_DENORMAL(%s);\n", src1);
+  ORC_ASM_CODE(p,"      _src2.i = ORC_DENORMAL(%s);\n", src2);
+  ORC_ASM_CODE(p,"      if (ORC_ISNAN(_src1.i)) %s = _src1.i;\n", dest);
+  ORC_ASM_CODE(p,"      else if (ORC_ISNAN(_src2.i)) %s = _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"      else %s = (_src1.f < _src2.f) ? _src1.i : _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"    }\n");
+}
+
+static void
+c_rule_maxf (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40], src2[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+  c_get_name_int (src2, p, insn, insn->src_args[1]);
+
+  ORC_ASM_CODE(p,"    {\n");
+  ORC_ASM_CODE(p,"      orc_union32 _src1;\n");
+  ORC_ASM_CODE(p,"      orc_union32 _src2;\n");
+  ORC_ASM_CODE(p,"      _src1.i = ORC_DENORMAL(%s);\n", src1);
+  ORC_ASM_CODE(p,"      _src2.i = ORC_DENORMAL(%s);\n", src2);
+  ORC_ASM_CODE(p,"      if (ORC_ISNAN(_src1.i)) %s = _src1.i;\n", dest);
+  ORC_ASM_CODE(p,"      else if (ORC_ISNAN(_src2.i)) %s = _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"      else %s = (_src1.f > _src2.f) ? _src1.i : _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"    }\n");
+}
+
+static void
+c_rule_mind (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40], src2[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+  c_get_name_int (src2, p, insn, insn->src_args[1]);
+
+  ORC_ASM_CODE(p,"    {\n");
+  ORC_ASM_CODE(p,"      orc_union64 _src1;\n");
+  ORC_ASM_CODE(p,"      orc_union64 _src2;\n");
+  ORC_ASM_CODE(p,"      _src1.i = ORC_DENORMAL_DOUBLE(%s);\n", src1);
+  ORC_ASM_CODE(p,"      _src2.i = ORC_DENORMAL_DOUBLE(%s);\n", src2);
+  ORC_ASM_CODE(p,"      if (ORC_ISNAN_DOUBLE(_src1.i)) %s = _src1.i;\n", dest);
+  ORC_ASM_CODE(p,"      else if (ORC_ISNAN_DOUBLE(_src2.i)) %s = _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"      else %s = (_src1.f < _src2.f) ? _src1.i : _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"    }\n");
+}
+
+static void
+c_rule_maxd (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src1[40], src2[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src1, p, insn, insn->src_args[0]);
+  c_get_name_int (src2, p, insn, insn->src_args[1]);
+
+  ORC_ASM_CODE(p,"    {\n");
+  ORC_ASM_CODE(p,"      orc_union64 _src1;\n");
+  ORC_ASM_CODE(p,"      orc_union64 _src2;\n");
+  ORC_ASM_CODE(p,"      _src1.i = ORC_DENORMAL_DOUBLE(%s);\n", src1);
+  ORC_ASM_CODE(p,"      _src2.i = ORC_DENORMAL_DOUBLE(%s);\n", src2);
+  ORC_ASM_CODE(p,"      if (ORC_ISNAN_DOUBLE(_src1.i)) %s = _src1.i;\n", dest);
+  ORC_ASM_CODE(p,"      else if (ORC_ISNAN_DOUBLE(_src2.i)) %s = _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"      else %s = (_src1.f > _src2.f) ? _src1.i : _src2.i;\n", dest);
+  ORC_ASM_CODE(p,"    }\n");
+}
+
 static OrcTarget c_target = {
   "c",
   FALSE,
@@ -964,6 +1305,7 @@ orc_c_init (void)
 #define BINARY_SQ(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_UQ(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_F(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define BINARY_D(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_SB(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_UB(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_SW(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
@@ -973,6 +1315,7 @@ orc_c_init (void)
 #define UNARY_SQ(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_UQ(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_F(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define UNARY_D(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_BW(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_WL(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_LW(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
@@ -985,15 +1328,20 @@ orc_c_init (void)
 #define UNARY_WB(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 
 #define UNARY_FL(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define UNARY_DL(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_FL(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define BINARY_DQ(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_LF(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define UNARY_LD(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define UNARY_DF(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define UNARY_FD(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 
 #include "opcodes.h"
 
-  orc_rule_register (rule_set, "loadpb", c_rule_loadpX, NULL);
-  orc_rule_register (rule_set, "loadpw", c_rule_loadpX, NULL);
-  orc_rule_register (rule_set, "loadpl", c_rule_loadpX, NULL);
-  orc_rule_register (rule_set, "loadpq", c_rule_loadpX, NULL);
+  orc_rule_register (rule_set, "loadpb", c_rule_loadpX, (void *)1);
+  orc_rule_register (rule_set, "loadpw", c_rule_loadpX, (void *)2);
+  orc_rule_register (rule_set, "loadpl", c_rule_loadpX, (void *)4);
+  orc_rule_register (rule_set, "loadpq", c_rule_loadpX, (void *)8);
   orc_rule_register (rule_set, "loadb", c_rule_loadX, NULL);
   orc_rule_register (rule_set, "loadw", c_rule_loadX, NULL);
   orc_rule_register (rule_set, "loadl", c_rule_loadX, NULL);
@@ -1015,6 +1363,7 @@ orc_c_init (void)
   orc_rule_register (rule_set, "accw", c_rule_accw, NULL);
   orc_rule_register (rule_set, "accl", c_rule_accl, NULL);
   orc_rule_register (rule_set, "accsadubl", c_rule_accsadubl, NULL);
+  orc_rule_register (rule_set, "splitql", c_rule_splitql, NULL);
   orc_rule_register (rule_set, "splitlw", c_rule_splitlw, NULL);
   orc_rule_register (rule_set, "splitwb", c_rule_splitwb, NULL);
   orc_rule_register (rule_set, "splatbw", c_rule_splatbw, NULL);
@@ -1022,5 +1371,15 @@ orc_c_init (void)
   orc_rule_register (rule_set, "splatw3q", c_rule_splatw3q, NULL);
   orc_rule_register (rule_set, "div255w", c_rule_div255w, NULL);
   orc_rule_register (rule_set, "divluw", c_rule_divluw, NULL);
+  orc_rule_register (rule_set, "convlf", c_rule_convlf, NULL);
+  orc_rule_register (rule_set, "convld", c_rule_convld, NULL);
+  orc_rule_register (rule_set, "convfl", c_rule_convfl, NULL);
+  orc_rule_register (rule_set, "convdl", c_rule_convdl, NULL);
+  orc_rule_register (rule_set, "convfd", c_rule_convfd, NULL);
+  orc_rule_register (rule_set, "convdf", c_rule_convdf, NULL);
+  orc_rule_register (rule_set, "minf", c_rule_minf, NULL);
+  orc_rule_register (rule_set, "maxf", c_rule_maxf, NULL);
+  orc_rule_register (rule_set, "mind", c_rule_mind, NULL);
+  orc_rule_register (rule_set, "maxd", c_rule_maxd, NULL);
 }
 
