@@ -96,7 +96,12 @@ static OrcTarget neon_target = {
   ORC_VEC_REG_BASE,
   orc_compiler_neon_get_default_flags,
   orc_compiler_neon_init,
-  orc_compiler_neon_assemble
+  orc_compiler_neon_assemble,
+  { { 0 } }, 0,
+  NULL,
+  NULL,
+  NULL,
+  orc_arm_flush_cache
 };
 
 void
@@ -174,6 +179,7 @@ orc_compiler_neon_init (OrcCompiler *compiler)
   }
 
   switch (orc_program_get_max_array_size (compiler->program)) {
+    case 0:
     case 1:
       loop_shift = 4;
       break;
@@ -187,8 +193,8 @@ orc_compiler_neon_init (OrcCompiler *compiler)
       loop_shift = 1;
       break;
     default:
-      ORC_ERROR("unhandled max var size %d",
-          orc_program_get_max_var_size (compiler->program));
+      ORC_ERROR("unhandled max array size %d",
+          orc_program_get_max_array_size (compiler->program));
       break;
   }
   if (loop_shift < compiler->loop_shift) {
@@ -212,8 +218,8 @@ orc_compiler_neon_init (OrcCompiler *compiler)
       loop_shift = 0;
       break;
     default:
-      ORC_ERROR("unhandled max var size %d",
-          orc_program_get_max_var_size (compiler->program));
+      ORC_ERROR("unhandled max accumulator size %d",
+          orc_program_get_max_accumulator_size (compiler->program));
       break;
   }
   if (loop_shift < compiler->loop_shift) {
@@ -472,6 +478,8 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
   int i;
   
   align_var = get_align_var (compiler);
+  if (compiler->error) return;
+
   var_size_shift = get_shift (compiler->vars[align_var].size);
   align_shift = 4;
 
@@ -700,8 +708,6 @@ orc_compiler_neon_assemble (OrcCompiler *compiler)
   orc_arm_emit_data (compiler, 0x0f0e0f0e);
 
   orc_arm_do_fixups (compiler);
-
-  orc_arm_flush_cache (compiler);
 }
 
 void
