@@ -47,10 +47,13 @@ static OrcTarget sse_target = {
   0,
   NULL,
   sse_load_constant,
-  sse_get_flag_name
+  sse_get_flag_name,
+  NULL
 };
 
 
+extern int orc_x86_sse_flags;
+extern int orc_x86_mmx_flags;
 
 void
 orc_sse_init (void)
@@ -62,11 +65,11 @@ orc_sse_init (void)
 
 #if defined(HAVE_I386)
 #ifndef MMX
-  if (!(orc_sse_get_cpu_flags () & ORC_TARGET_SSE_SSE2)) {
+  if (!(orc_x86_sse_flags & ORC_TARGET_SSE_SSE2)) {
     sse_target.executable = FALSE;
   }
 #else
-  if (!(orc_mmx_get_cpu_flags () & ORC_TARGET_MMX_MMX)) {
+  if (!(orc_x86_mmx_flags & ORC_TARGET_MMX_MMX)) {
     mmx_target.executable = FALSE;
   }
 #endif
@@ -90,7 +93,11 @@ orc_compiler_sse_get_default_flags (void)
   }
   
 #if defined(HAVE_AMD64) || defined(HAVE_I386)
-  flags |= orc_sse_get_cpu_flags ();
+#ifndef MMX
+  flags |= orc_x86_sse_flags;
+#else
+  flags |= orc_x86_mmx_flags;
+#endif
 #else
 #ifndef MMX
   flags |= ORC_TARGET_SSE_SSE2;
@@ -145,7 +152,6 @@ orc_compiler_sse_init (OrcCompiler *compiler)
     for(i=ORC_GP_REG_BASE;i<ORC_GP_REG_BASE+16;i++){
       compiler->valid_regs[i] = 1;
     }
-    compiler->valid_regs[X86_EDI] = 0;
     compiler->valid_regs[X86_ESP] = 0;
 #ifndef MMX
     for(i=X86_XMM0;i<X86_XMM0+16;i++){
@@ -162,6 +168,13 @@ orc_compiler_sse_init (OrcCompiler *compiler)
     compiler->save_regs[X86_R13] = 1;
     compiler->save_regs[X86_R14] = 1;
     compiler->save_regs[X86_R15] = 1;
+#ifdef HAVE_OS_WIN32
+    compiler->save_regs[X86_EDI] = 1;
+    compiler->save_regs[X86_ESI] = 1;
+    for(i=X86_XMM0+6;i<X86_XMM0+16;i++){
+      compiler->save_regs[i] = 1;
+    }
+#endif
   } else {
     for(i=ORC_GP_REG_BASE;i<ORC_GP_REG_BASE+8;i++){
       compiler->valid_regs[i] = 1;
@@ -186,7 +199,12 @@ orc_compiler_sse_init (OrcCompiler *compiler)
   compiler->valid_regs[compiler->gp_tmpreg] = 0;
 
   if (compiler->is_64bit) {
+#ifdef HAVE_OS_WIN32
+    compiler->exec_reg = X86_ECX;
+    compiler->gp_tmpreg = X86_EAX;
+#else
     compiler->exec_reg = X86_EDI;
+#endif
   } else {
     if (compiler->use_frame_pointer) {
       compiler->exec_reg = X86_EBX;
@@ -664,7 +682,7 @@ orc_compiler_sse_assemble (OrcCompiler *compiler)
   {
     orc_sse_emit_loop (compiler, 0, 0);
 
-    compiler->codeptr = compiler->program->code;
+    compiler->codeptr = compiler->code;
     free (compiler->asm_code);
     compiler->asm_code = NULL;
     compiler->asm_code_len = 0;
@@ -672,6 +690,8 @@ orc_compiler_sse_assemble (OrcCompiler *compiler)
     compiler->n_fixups = 0;
   }
 
+  if (compiler->error) return;
+
   orc_x86_emit_prologue (compiler);
 
 #ifndef MMX
