@@ -147,13 +147,18 @@ orc_program_free (OrcProgram *program)
 {
   int i;
   for(i=0;i<ORC_N_VARIABLES;i++){
-    if (program->vars[i].name) free (program->vars[i].name);
+    if (program->vars[i].name) {
+      free (program->vars[i].name);
+      program->vars[i].name = NULL;
+    }
   }
   if (program->asm_code) {
     free (program->asm_code);
+    program->asm_code = NULL;
   }
   if (program->name) {
     free (program->name);
+    program->name = NULL;
   }
   free (program);
 }
@@ -345,7 +350,95 @@ orc_program_add_constant (OrcProgram *program, int size, int value, const char *
 
   program->vars[i].vartype = ORC_VAR_TYPE_CONST;
   program->vars[i].size = size;
-  program->vars[i].value = value;
+  program->vars[i].value.i = value;
+  program->vars[i].name = strdup(name);
+  program->n_const_vars++;
+
+  return i;
+}
+
+int
+orc_program_add_constant_int64 (OrcProgram *program, int size,
+    orc_int64 value, const char *name)
+{
+  int i;
+  
+  i = ORC_VAR_C1 + program->n_const_vars;
+
+  program->vars[i].vartype = ORC_VAR_TYPE_CONST;
+  program->vars[i].size = size;
+  program->vars[i].value.i = value;
+  program->vars[i].name = strdup(name);
+  program->n_const_vars++;
+
+  return i;
+}
+
+int
+orc_program_add_constant_float (OrcProgram *program, int size,
+    float value, const char *name)
+{
+  orc_union32 u;
+  u.f = value;
+  return orc_program_add_constant (program, size, u.i, name);
+}
+
+int
+orc_program_add_constant_double (OrcProgram *program, int size,
+    double value, const char *name)
+{
+  orc_union64 u;
+  u.f = value;
+  return orc_program_add_constant_int64 (program, size, u.i, name);
+}
+
+int
+orc_program_add_constant_str (OrcProgram *program, int size,
+    const char *value, const char *name)
+{
+  int i;
+  char *end;
+  orc_int64 val_i;
+  double val_d;
+  int j;
+
+  i = ORC_VAR_C1 + program->n_const_vars;
+
+  val_i = _strtoll (value, &end, 0);
+  if (end[0] == 0) {
+    program->vars[i].value.i = val_i;
+    if (size == 0)
+      size = 4;
+  } else if ((end[0] == 'l' || end[0] == 'L') && end[1] == 0) {
+    program->vars[i].value.i = val_i;
+    if (size == 0)
+      size = 8;
+  } else {
+    val_d = strtod (value, &end);
+
+    if (end[0] == 0) {
+      orc_union32 u;
+      u.f = val_d;
+      program->vars[i].value.i = u.i;
+      if (size == 0)
+        size = 4;
+    } else if ((end[0] == 'l' || end[0] == 'L') && end[1] == 0) {
+      program->vars[i].value.f = val_d;
+      if (size == 0)
+        size = 8;
+    } else {
+      return -1;
+    }
+  }
+
+  for(j=0;j<program->n_const_vars;j++){
+    if (program->vars[ORC_VAR_C1 + j].value.i == program->vars[i].value.i) {
+      return ORC_VAR_C1 + j;
+    }
+  }
+
+  program->vars[i].vartype = ORC_VAR_TYPE_CONST;
+  program->vars[i].size = size;
   program->vars[i].name = strdup(name);
   program->n_const_vars++;
 
@@ -368,6 +461,7 @@ orc_program_add_parameter (OrcProgram *program, int size, const char *name)
   int i = ORC_VAR_P1 + program->n_param_vars;
 
   program->vars[i].vartype = ORC_VAR_TYPE_PARAM;
+  program->vars[i].param_type = ORC_PARAM_TYPE_INT;
   program->vars[i].size = size;
   program->vars[i].name = strdup(name);
   program->n_param_vars++;
@@ -391,7 +485,37 @@ orc_program_add_parameter_float (OrcProgram *program, int size, const char *name
   int i = ORC_VAR_P1 + program->n_param_vars;
 
   program->vars[i].vartype = ORC_VAR_TYPE_PARAM;
-  program->vars[i].is_float_param = TRUE;
+  program->vars[i].param_type = ORC_PARAM_TYPE_FLOAT;
+  program->vars[i].size = size;
+  program->vars[i].name = strdup(name);
+  program->n_param_vars++;
+
+  return i;
+}
+
+int
+orc_program_add_parameter_double (OrcProgram *program, int size,
+    const char *name)
+{
+  int i = ORC_VAR_P1 + program->n_param_vars;
+
+  program->vars[i].vartype = ORC_VAR_TYPE_PARAM;
+  program->vars[i].param_type = ORC_PARAM_TYPE_DOUBLE;
+  program->vars[i].size = size;
+  program->vars[i].name = strdup(name);
+  program->n_param_vars++;
+
+  return i;
+}
+
+int
+orc_program_add_parameter_int64 (OrcProgram *program, int size,
+    const char *name)
+{
+  int i = ORC_VAR_P1 + program->n_param_vars;
+
+  program->vars[i].vartype = ORC_VAR_TYPE_PARAM;
+  program->vars[i].param_type = ORC_PARAM_TYPE_INT64;
   program->vars[i].size = size;
   program->vars[i].name = strdup(name);
   program->n_param_vars++;
@@ -799,6 +923,10 @@ orc_program_get_max_accumulator_size (OrcProgram *program)
 int _orc_data_cache_size_level1;
 int _orc_data_cache_size_level2;
 int _orc_data_cache_size_level3;
+int _orc_cpu_family;
+int _orc_cpu_model;
+int _orc_cpu_stepping;
+const char *_orc_cpu_name = "unknown";
 
 void
 orc_get_data_cache_sizes (int *level1, int *level2, int *level3)
@@ -815,3 +943,44 @@ orc_get_data_cache_sizes (int *level1, int *level2, int *level3)
 
 }
 
+void
+orc_get_cpu_family_model_stepping (int *family, int *model, int *stepping)
+{
+  if (family) {
+    *family = _orc_cpu_family;
+  }
+  if (model) {
+    *model = _orc_cpu_model;
+  }
+  if (stepping) {
+    *stepping = _orc_cpu_stepping;
+  }
+}
+
+const char *
+orc_get_cpu_name (void)
+{
+  return _orc_cpu_name;
+}
+
+void
+orc_program_reset (OrcProgram *program)
+{
+  if (program->orccode) {
+    orc_code_free (program->orccode);
+    program->orccode = NULL;
+  }
+  if (program->asm_code) {
+    free(program->asm_code);
+    program->asm_code = NULL;
+  }
+}
+
+OrcCode *
+orc_program_take_code (OrcProgram *program)
+{
+  OrcCode *code = program->orccode;
+  program->orccode = NULL;
+  return code;
+}
+
