@@ -19,6 +19,7 @@ typedef struct _OrcConstant OrcConstant;
 typedef struct _OrcFixup OrcFixup;
 typedef struct _OrcTarget OrcTarget;
 typedef struct _OrcCode OrcCode;
+typedef struct _OrcCodeChunk OrcCodeChunk;
 
 typedef void (*OrcOpcodeEmulateFunc)(OrcOpcodeExecutor *ex, void *user);
 typedef void (*OrcOpcodeEmulateNFunc)(OrcOpcodeExecutor *ex, int index, int n);
@@ -156,7 +157,8 @@ enum {
   ORC_VAR_T12,
   ORC_VAR_T13,
   ORC_VAR_T14,
-  ORC_VAR_T15
+  ORC_VAR_T15,
+  ORC_VAR_T16
 };
 
 enum {
@@ -175,6 +177,13 @@ enum {
   ORC_SAMPLE_FOUR_TAP
 };
 
+enum {
+  ORC_PARAM_TYPE_INT = 0,
+  ORC_PARAM_TYPE_FLOAT,
+  ORC_PARAM_TYPE_INT64,
+  ORC_PARAM_TYPE_DOUBLE
+};
+
 typedef enum {
   ORC_COMPILE_RESULT_OK = 0,
 
@@ -214,13 +223,13 @@ struct _OrcVariable {
   int is_aligned;
   int is_uncached;
 
-  int value;
+  orc_union64 value;
 
   int ptr_register;
   int ptr_offset;
   int mask_alloc;
   int aligned_data;
-  int is_float_param;
+  int param_type;
   int load_dest;
 };
 
@@ -346,7 +355,34 @@ struct _OrcProgram {
   } _unused[ORC_N_INSNS]; /* needed for ABI compatibility */
   int n_insns;
 
-  OrcVariable vars[ORC_N_VARIABLES];
+  struct {
+    char *name;
+    char *type_name;
+
+    int size;
+    OrcVarType vartype;
+
+    int used;
+    int first_use;
+    int last_use;
+    int replaced;
+    int replacement;
+
+    int alloc;
+    int is_chained;
+    int is_aligned;
+    int is_uncached;
+
+    int value;
+
+    int ptr_register;
+    int ptr_offset;
+    int mask_alloc;
+    int aligned_data;
+    int param_type;
+    int load_dest;
+  } _unused3[ORC_N_VARIABLES]; /* needed for ABI compatibility */
+
   int n_src_vars;
   int n_dest_vars;
   int n_param_vars;
@@ -357,11 +393,12 @@ struct _OrcProgram {
   char *name;
   char *asm_code;
 
-  unsigned char *code;
+  unsigned char *_unused2;
+  /* The offset of code_exec in this structure is part of the ABI */
   void *code_exec;
-  int code_size;
 
   OrcInstruction insns[ORC_N_INSNS];
+  OrcVariable vars[ORC_N_VARIABLES];
 
   void *backup_func;
   int is_2d;
@@ -393,6 +430,7 @@ struct _OrcCompiler {
   int n_temp_vars;
   int n_dup_vars;
 
+  unsigned char *code;
   unsigned char *codeptr;
   
   OrcConstant constants[ORC_N_CONSTANTS];
@@ -445,8 +483,8 @@ struct _OrcCompiler {
 #define ORC_DEST_ARG(p,i,n) ((p)->vars[(i)->dest_args[(n)]].alloc)
 #define ORC_SRC_TYPE(p,i,n) ((p)->vars[(i)->src_args[(n)]].vartype)
 #define ORC_DEST_TYPE(p,i,n) ((p)->vars[(i)->dest_args[(n)]].vartype)
-#define ORC_SRC_VAL(p,i,n) ((p)->vars[(i)->src_args[(n)]].value)
-#define ORC_DEST_VAL(p,i,n) ((p)->vars[(i)->dest_args[(n)]].value)
+#define ORC_SRC_VAL(p,insn,n) ((p)->vars[(insn)->src_args[(n)]].value.i)
+#define ORC_DEST_VAL(p,insn,n) ((p)->vars[(insn)->dest_args[(n)]].value.i)
 
 /**
  * OrcOpcodeExecutor:
@@ -486,7 +524,7 @@ struct _OrcExecutor {
   /* m is stored in params[ORC_VAR_A1] */
   /* m_index is stored in params[ORC_VAR_A2] */
   /* elapsed time is stored in params[ORC_VAR_A3] */
-  /* source resampling parameters are in params[ORC_VAR_C1..C8] */
+  /* high half of params is stored in params[ORC_VAR_T1..] */
 };
 
 /* the alternate view of OrcExecutor */
@@ -506,9 +544,10 @@ struct _OrcExecutorAlt {
   int m_index;
   int time;
   int unused2;
-  int src_resample[8];
+  int unused4[8];
   int params[ORC_VAR_T1-ORC_VAR_P1];
-  int unused3[ORC_N_VARIABLES - ORC_VAR_T1];
+  int params_hi[ORC_VAR_T1-ORC_VAR_P1];
+  int unused3[ORC_N_VARIABLES - ORC_VAR_T9];
   int accumulators[4];
 };
 #define ORC_EXECUTOR_EXEC(ex) ((OrcExecutorFunc)((ex)->arrays[ORC_VAR_A1]))
@@ -516,6 +555,14 @@ struct _OrcExecutorAlt {
 #define ORC_EXECUTOR_M_INDEX(ex) ((ex)->params[ORC_VAR_A2])
 #define ORC_EXECUTOR_TIME(ex) ((ex)->params[ORC_VAR_A3])
 
+typedef struct _OrcCodeVariable OrcCodeVariable;
+struct _OrcCodeVariable {
+  /*< private >*/
+  int vartype;
+  int size;
+  orc_union64 value;
+};
+
 struct _OrcCode {
   /*< private >*/
   OrcCompileResult result;
@@ -525,11 +572,12 @@ struct _OrcCode {
   OrcExecutorFunc exec;
   unsigned char *code;
   int code_size;
+  void *chunk;
 
   /* for emulation */
   int n_insns;
   OrcInstruction *insns;
-  OrcVariable *vars;
+  OrcCodeVariable *vars;
   int is_2d;
   int constant_n;
   int constant_m;
@@ -554,8 +602,9 @@ struct _OrcTarget {
   const char * (*get_asm_preamble)(void);
   void (*load_constant)(OrcCompiler *compiler, int reg, int size, int value);
   const char * (*get_flag_name)(int shift);
+  void (*flush_cache) (OrcCode *code);
 
-  void *_unused[7];
+  void *_unused[6];
 };
 
 
@@ -613,8 +662,14 @@ int orc_program_dup_temporary (OrcProgram *program, int i, int j);
 int orc_program_add_source (OrcProgram *program, int size, const char *name);
 int orc_program_add_destination (OrcProgram *program, int size, const char *name);
 int orc_program_add_constant (OrcProgram *program, int size, int value, const char *name);
+int orc_program_add_constant_int64 (OrcProgram *program, int size, orc_int64 value, const char *name);
+int orc_program_add_constant_float (OrcProgram *program, int size, float value, const char *name);
+int orc_program_add_constant_double (OrcProgram *program, int size, double value, const char *name);
+int orc_program_add_constant_str (OrcProgram *program, int size, const char *value, const char *name);
 int orc_program_add_parameter (OrcProgram *program, int size, const char *name);
 int orc_program_add_parameter_float (OrcProgram *program, int size, const char *name);
+int orc_program_add_parameter_double (OrcProgram *program, int size, const char *name);
+int orc_program_add_parameter_int64 (OrcProgram *program, int size, const char *name);
 int orc_program_add_accumulator (OrcProgram *program, int size, const char *name);
 void orc_program_set_type_name (OrcProgram *program, int var, const char *type_name);
 void orc_program_set_sampling_type (OrcProgram *program, int var, int sampling_type);
@@ -627,6 +682,9 @@ void orc_executor_set_stride (OrcExecutor *ex, int var, int stride);
 void orc_executor_set_array_str (OrcExecutor *ex, const char *name, void *ptr);
 void orc_executor_set_param (OrcExecutor *ex, int var, int value);
 void orc_executor_set_param_str (OrcExecutor *ex, const char *name, int value);
+void orc_executor_set_param_float (OrcExecutor *ex, int var, float value);
+void orc_executor_set_param_int64 (OrcExecutor *ex, int var, orc_int64 value);
+void orc_executor_set_param_double (OrcExecutor *ex, int var, double value);
 int orc_executor_get_accumulator (OrcExecutor *ex, int var);
 int orc_executor_get_accumulator_str (OrcExecutor *ex, const char *name);
 void orc_executor_set_n (OrcExecutor *ex, int n);
@@ -653,13 +711,15 @@ const char * orc_target_get_flag_name (OrcTarget *target, int shift);
 
 int orc_program_allocate_register (OrcProgram *program, int is_data);
 
-void orc_compiler_allocate_codemem (OrcCompiler *compiler);
+void orc_code_allocate_codemem (OrcCode *code, int size);
 int orc_compiler_label_new (OrcCompiler *compiler);
 int orc_compiler_get_constant (OrcCompiler *compiler, int size, int value);
 int orc_compiler_get_temp_constant (OrcCompiler *compiler, int size, int value);
 int orc_compiler_get_temp_reg (OrcCompiler *compiler);
 int orc_compiler_get_constant_reg (OrcCompiler *compiler);
 
+void orc_program_reset (OrcProgram *program);
+OrcCode *orc_program_take_code (OrcProgram *program);
 
 const char *orc_program_get_asm_code (OrcProgram *program);
 const char *orc_target_get_asm_preamble (const char *target);
@@ -676,6 +736,8 @@ int orc_program_get_max_array_size (OrcProgram *program);
 int orc_program_get_max_accumulator_size (OrcProgram *program);
 
 void orc_get_data_cache_sizes (int *level1, int *level2, int *level3);
+void orc_get_cpu_family_model_stepping (int *family, int *model, int *stepping);
+const char * orc_get_cpu_name (void);
 
 OrcCode * orc_code_new (void);
 void orc_code_free (OrcCode *code);
@@ -688,10 +750,16 @@ int orc_compiler_flag_check (const char *flag);
 extern int _orc_data_cache_size_level1;
 extern int _orc_data_cache_size_level2;
 extern int _orc_data_cache_size_level3;
+extern int _orc_cpu_family;
+extern int _orc_cpu_model;
+extern int _orc_cpu_stepping;
+extern const char *_orc_cpu_name;
 
 extern int _orc_compiler_flag_backup;
 extern int _orc_compiler_flag_debug;
 
+void orc_code_chunk_free (OrcCodeChunk *chunk);
+
 #endif
 
 #endif
