@@ -196,11 +196,12 @@ orc_x86_emit_mov_memoffset_reg (OrcCompiler *compiler, int size, int offset,
 
   switch (size) {
     case 1:
-      ORC_ASM_CODE(compiler,"  movb %d(%%%s), %%%s\n", offset,
+      ORC_ASM_CODE(compiler,"  movzx %d(%%%s), %%%s\n", offset,
           orc_x86_get_regname_ptr(compiler, reg1),
-          orc_x86_get_regname_8(reg2));
+          orc_x86_get_regname(reg2));
       orc_x86_emit_rex(compiler, size, reg2, 0, reg1);
-      *compiler->codeptr++ = 0x8a;
+      *compiler->codeptr++ = 0x0f;
+      *compiler->codeptr++ = 0xb6;
       orc_x86_emit_modrm_memoffset (compiler, reg2, offset, reg1);
       return;
     case 2:
@@ -424,7 +425,7 @@ orc_x86_emit_add_imm_memoffset (OrcCompiler *compiler, int size, int value,
     ORC_ASM_CODE(compiler,"  addl $%d, %d(%%%s)\n", value, offset,
         orc_x86_get_regname_ptr(compiler, reg));
   } else {
-    ORC_ASM_CODE(compiler,"  add $%d, %d(%%%s)\n", value, offset,
+    ORC_ASM_CODE(compiler,"  addq $%d, %d(%%%s)\n", value, offset,
         orc_x86_get_regname_ptr(compiler, reg));
   }
 
@@ -438,7 +439,7 @@ orc_x86_emit_add_imm_memoffset (OrcCompiler *compiler, int size, int value,
     orc_x86_emit_modrm_memoffset (compiler, 0, offset, reg);
     *compiler->codeptr++ = (value & 0xff);
     *compiler->codeptr++ = ((value>>8) & 0xff);
-    if (size == 4) {
+    if (size == 4 || size == 8) {
       *compiler->codeptr++ = ((value>>16) & 0xff);
       *compiler->codeptr++ = ((value>>24) & 0xff);
     }
@@ -497,7 +498,7 @@ orc_x86_emit_add_imm_reg (OrcCompiler *compiler, int size, int value, int reg, o
   } else if (size == 4) {
     ORC_ASM_CODE(compiler,"  addl $%d, %%%s\n", value, orc_x86_get_regname(reg));
   } else {
-    ORC_ASM_CODE(compiler,"  add $%d, %%%s\n", value, orc_x86_get_regname_64(reg));
+    ORC_ASM_CODE(compiler,"  addq $%d, %%%s\n", value, orc_x86_get_regname_64(reg));
   }
 
   orc_x86_emit_rex(compiler, size, 0, 0, reg);
@@ -510,7 +511,7 @@ orc_x86_emit_add_imm_reg (OrcCompiler *compiler, int size, int value, int reg, o
     orc_x86_emit_modrm_reg (compiler, reg, 0);
     *compiler->codeptr++ = (value & 0xff);
     *compiler->codeptr++ = ((value>>8) & 0xff);
-    if (size == 4) {
+    if (size == 4 || size == 8) {
       *compiler->codeptr++ = ((value>>16) & 0xff);
       *compiler->codeptr++ = ((value>>24) & 0xff);
     }
@@ -998,7 +999,7 @@ orc_x86_emit_align (OrcCompiler *compiler)
   int diff;
   int align_shift = 4;
 
-  diff = (compiler->program->code - compiler->codeptr)&((1<<align_shift) - 1);
+  diff = (compiler->code - compiler->codeptr)&((1<<align_shift) - 1);
   while (diff) {
     ORC_ASM_CODE(compiler,"  nop\n");
     *compiler->codeptr++ = 0x90;
