@@ -2,7 +2,6 @@
 #include "config.h"
 
 #include <orc/orc.h>
-#include <orc-test/orctest.h>
 #include <orc/orcparse.h>
 
 #include <stdio.h>
@@ -91,7 +90,6 @@ main (int argc, char *argv[])
   char *log = NULL;
 
   orc_init ();
-  orc_test_init ();
 
   for(i=1;i<argc;i++) {
     if (strcmp(argv[i], "--header") == 0) {
@@ -473,7 +471,10 @@ const char *enumnames[] = {
   "ORC_VAR_T1", "ORC_VAR_T2", "ORC_VAR_T3", "ORC_VAR_T4",
   "ORC_VAR_T5", "ORC_VAR_T6", "ORC_VAR_T7", "ORC_VAR_T8",
   "ORC_VAR_T9", "ORC_VAR_T10", "ORC_VAR_T11", "ORC_VAR_T12",
-  "ORC_VAR_T13", "ORC_VAR_T14", "ORC_VAR_T15", "ORC_VAR_T16"
+  "ORC_VAR_T13", "ORC_VAR_T14", "ORC_VAR_T15",
+  /* ORC_VAR_T16 is a special case because it didn't exist in
+     the API until 0.4.8 */
+  "47"
 };
 
 void
@@ -537,12 +538,25 @@ output_prototype (OrcProgram *p, FILE *output)
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
       if (need_comma) fprintf(output, ", ");
-      if (var->is_float_param) {
-        REQUIRE(0,4,5,1);
+      switch (var->param_type) {
+        case ORC_PARAM_TYPE_INT:
+          fprintf(output, "int %s", varnames[ORC_VAR_P1 + i]);
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          REQUIRE(0,4,5,1);
+          fprintf(output, "float %s", varnames[ORC_VAR_P1 + i]);
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          REQUIRE(0,4,7,1);
+          fprintf(output, "orc_int64 %s", varnames[ORC_VAR_P1 + i]);
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          REQUIRE(0,4,7,1);
+          fprintf(output, "double %s", varnames[ORC_VAR_P1 + i]);
+          break;
+        default:
+          ORC_ASSERT(0);
       }
-      fprintf(output, "%s %s",
-          var->is_float_param ? "float" : "int",
-          varnames[ORC_VAR_P1 + i]);
       need_comma = TRUE;
     }
   }
@@ -571,7 +585,11 @@ output_code_backup (OrcProgram *p, FILE *output)
 {
 
   fprintf(output, "static void\n");
-  fprintf(output, "_backup_%s (OrcExecutor *ex)\n", p->name);
+  if (compat < ORC_VERSION(0,4,7,1)) {
+    fprintf(output, "_backup_%s (OrcExecutor * ex)\n", p->name);
+  } else {
+    fprintf(output, "_backup_%s (OrcExecutor * ORC_RESTRICT ex)\n", p->name);
+  }
   fprintf(output, "{\n");
   {
     OrcCompileResult result;
@@ -715,17 +733,44 @@ output_code_execute (OrcProgram *p, FILE *output, int is_inline)
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
-      if (var->is_float_param) {
-        REQUIRE(0,4,5,1);
-        fprintf(output, "  {\n");
-        fprintf(output, "    orc_union32 tmp;\n");
-        fprintf(output, "    tmp.f = %s;\n", varnames[ORC_VAR_P1 + i]);
-        fprintf(output, "    ex->params[%s] = tmp.i;\n",
-            enumnames[ORC_VAR_P1 + i]);
-        fprintf(output, "  }\n");
-      } else {
-        fprintf(output, "  ex->params[%s] = %s;\n",
-            enumnames[ORC_VAR_P1 + i], varnames[ORC_VAR_P1 + i]);
+      switch (var->param_type) {
+        case ORC_PARAM_TYPE_INT:
+          fprintf(output, "  ex->params[%s] = %s;\n",
+              enumnames[ORC_VAR_P1 + i], varnames[ORC_VAR_P1 + i]);
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          REQUIRE(0,4,5,1);
+          fprintf(output, "  {\n");
+          fprintf(output, "    orc_union32 tmp;\n");
+          fprintf(output, "    tmp.f = %s;\n", varnames[ORC_VAR_P1 + i]);
+          fprintf(output, "    ex->params[%s] = tmp.i;\n",
+              enumnames[ORC_VAR_P1 + i]);
+          fprintf(output, "  }\n");
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          REQUIRE(0,4,7,1);
+          fprintf(output, "  {\n");
+          fprintf(output, "    orc_union64 tmp;\n");
+          fprintf(output, "    tmp.i = %s;\n", varnames[ORC_VAR_P1 + i]);
+          fprintf(output, "    ex->params[%s] = tmp.x2[0];\n",
+              enumnames[ORC_VAR_P1 + i]);
+          fprintf(output, "    ex->params[%s] = tmp.x2[1];\n",
+              enumnames[ORC_VAR_T1 + i]);
+          fprintf(output, "  }\n");
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          REQUIRE(0,4,5,1);
+          fprintf(output, "  {\n");
+          fprintf(output, "    orc_union64 tmp;\n");
+          fprintf(output, "    tmp.f = %s;\n", varnames[ORC_VAR_P1 + i]);
+          fprintf(output, "    ex->params[%s] = tmp.x2[0];\n",
+              enumnames[ORC_VAR_P1 + i]);
+          fprintf(output, "    ex->params[%s] = tmp.x2[1];\n",
+              enumnames[ORC_VAR_T1 + i]);
+          fprintf(output, "  }\n");
+          break;
+        default:
+          ORC_ASSERT(0);
       }
     }
   }
@@ -789,20 +834,41 @@ output_program_generation (OrcProgram *p, FILE *output, int is_inline)
   }
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_C1 + i];
-    if (var->size) {
+    if (var->size <= 4) {
       fprintf(output, "      orc_program_add_constant (p, %d, 0x%08x, \"%s\");\n",
-          var->size, var->value, varnames[ORC_VAR_C1 + i]);
+          var->size, (int)var->value.i, varnames[ORC_VAR_C1 + i]);
+    } else if (var->size > 4) {
+      REQUIRE(0,4,8,1);
+      fprintf(output, "      orc_program_add_constant_int64 (p, %d, 0x%08x%08x, \"%s\");\n",
+          var->size, (orc_uint32)(((orc_uint64)var->value.i)>>32),
+          (orc_uint32)(var->value.i), varnames[ORC_VAR_C1 + i]);
     }
   }
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
-      if (var->is_float_param) {
-        REQUIRE(0,4,5,1);
+      const char *suffix = NULL;
+      switch (var->param_type) {
+        case ORC_PARAM_TYPE_INT:
+          suffix="";
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          REQUIRE(0,4,5,1);
+          suffix="_float";
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          REQUIRE(0,4,7,1);
+          suffix="_int64";
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          REQUIRE(0,4,7,1);
+          suffix="_double";
+          break;
+        default:
+          ORC_ASSERT(0);
       }
       fprintf(output, "      orc_program_add_parameter%s (p, %d, \"%s\");\n",
-          var->is_float_param ? "_float" : "",
-          var->size, varnames[ORC_VAR_P1 + i]);
+          suffix, var->size, varnames[ORC_VAR_P1 + i]);
     }
   }
   for(i=0;i<16;i++){
@@ -938,24 +1004,40 @@ output_code_test (OrcProgram *p, FILE *output)
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_C1 + i];
     if (var->size) {
-      if (var->value != 0x80000000) {
+      if (var->value.i != 0x80000000) {
         fprintf(output, "      orc_program_add_constant (p, %d, %u, \"%s\");\n",
-            var->size, var->value, varnames[ORC_VAR_C1 + i]);
+            var->size, (int)var->value.i, varnames[ORC_VAR_C1 + i]);
       } else {
         fprintf(output, "      orc_program_add_constant (p, %d, 0x%08x, \"%s\");\n",
-            var->size, var->value, varnames[ORC_VAR_C1 + i]);
+            var->size, (int)var->value.i, varnames[ORC_VAR_C1 + i]);
       }
     }
   }
   for(i=0;i<8;i++){
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
-      if (var->is_float_param) {
-        REQUIRE(0,4,5,1);
+      const char *suffix = NULL;
+      switch (var->param_type) {
+        case ORC_PARAM_TYPE_INT:
+          suffix="";
+          break;
+        case ORC_PARAM_TYPE_FLOAT:
+          REQUIRE(0,4,5,1);
+          suffix="_float";
+          break;
+        case ORC_PARAM_TYPE_INT64:
+          REQUIRE(0,4,7,1);
+          suffix="_int64";
+          break;
+        case ORC_PARAM_TYPE_DOUBLE:
+          REQUIRE(0,4,7,1);
+          suffix="_double";
+          break;
+        default:
+          ORC_ASSERT(0);
       }
       fprintf(output, "    orc_program_add_parameter%s (p, %d, \"%s\");\n",
-          var->is_float_param ? "_float" : "",
-          var->size, varnames[ORC_VAR_P1 + i]);
+          suffix, var->size, varnames[ORC_VAR_P1 + i]);
     }
   }
   for(i=0;i<16;i++){
@@ -984,19 +1066,40 @@ output_code_test (OrcProgram *p, FILE *output)
             enumnames[insn->src_args[0]]);
       }
     } else {
-      if (p->vars[insn->src_args[1]].size != 0) {
-        fprintf(output, "      orc_program_append_2 (p, \"%s\", %d, %s, %s, %s, -1);\n",
-            insn->opcode->name, insn->flags, enumnames[insn->dest_args[0]],
-            enumnames[insn->src_args[0]], enumnames[insn->src_args[1]]);
-      } else {
-        fprintf(output, "      orc_program_append_2 (p, \"%s\", %d, %s, %s, -1, -1);\n",
-            insn->opcode->name, insn->flags, enumnames[insn->dest_args[0]],
-            enumnames[insn->src_args[0]]);
+      int args[4] = { 0, 0, 0, 0 };
+      int n_args = 0;
+
+      if (insn->opcode->dest_size[0] != 0) {
+        args[n_args++] = insn->dest_args[0];
       }
+      if (insn->opcode->dest_size[1] != 0) {
+        args[n_args++] = insn->dest_args[1];
+      }
+      if (insn->opcode->src_size[0] != 0) {
+        args[n_args++] = insn->src_args[0];
+      }
+      if (insn->opcode->src_size[1] != 0) {
+        args[n_args++] = insn->src_args[1];
+      }
+      if (insn->opcode->src_size[2] != 0) {
+        args[n_args++] = insn->src_args[2];
+      }
+
+      fprintf(output, "      orc_program_append_2 (p, \"%s\", %d, %s, %s, %s, %s);\n",
+          insn->opcode->name, insn->flags, enumnames[args[0]],
+          enumnames[args[1]], enumnames[args[2]],
+          enumnames[args[3]]);
     }
   }
 
   fprintf(output, "\n");
+  if (compat >= ORC_VERSION(0,4,7,1)) {
+    fprintf(output, "    if (benchmark) {\n");
+    fprintf(output, "      printf (\"    cycles (emulate) :   %%g\\n\",\n");
+    fprintf(output, "          orc_test_performance_full (p, ORC_TEST_FLAGS_EMULATE, NULL));\n");
+    fprintf(output, "    }\n");
+    fprintf(output, "\n");
+  }
   fprintf(output, "    ret = orc_test_compare_output_backup (p);\n");
   fprintf(output, "    if (!ret) {\n");
   fprintf(output, "      error = TRUE;\n");
@@ -1005,8 +1108,8 @@ output_code_test (OrcProgram *p, FILE *output)
   fprintf(output, "    }\n");
   fprintf(output, "\n");
   fprintf(output, "    if (benchmark) {\n");
-  fprintf(output, "      printf (\"    cycles (backup)  :   \");\n");
-  fprintf(output, "      orc_test_performance (p, ORC_TEST_FLAGS_BACKUP);\n");
+  fprintf(output, "      printf (\"    cycles (backup)  :   %%g\\n\",\n");
+  fprintf(output, "          orc_test_performance_full (p, ORC_TEST_FLAGS_BACKUP, NULL));\n");
   fprintf(output, "    }\n");
   fprintf(output, "\n");
   fprintf(output, "    ret = orc_test_compare_output (p);\n");
@@ -1019,8 +1122,8 @@ output_code_test (OrcProgram *p, FILE *output)
   fprintf(output, "    }\n");
   fprintf(output, "\n");
   fprintf(output, "    if (benchmark) {\n");
-  fprintf(output, "      printf (\"    cycles (compiled):   \");\n");
-  fprintf(output, "      orc_test_performance (p, 0);\n");
+  fprintf(output, "      printf (\"    cycles (compiled):   %%g\\n\",\n");
+  fprintf(output, "          orc_test_performance_full (p, 0, NULL));\n");
   fprintf(output, "    }\n");
   fprintf(output, "\n");
   fprintf(output, "    orc_program_free (p);\n");
